from dataclasses import dataclass, field
from typing import List, Union, Literal, Optional, Dict, Any

from podonos.common.validator import Rules, validate_args


@dataclass
class StimulusStatsFile:
    name: str
    model_tag: str
    tags: List[str]
    type: Union[Literal["A"], Literal["B"]]

    @staticmethod
    @validate_args(data=Rules.dict_not_none)
    def from_dict(data: Dict[str, Any]) -> "StimulusStatsFile":
        required_keys = ["name", "model_tag", "tags", "type"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Invalid data format for StimulusStatsFile: {data}")
        return StimulusStatsFile(name=data["name"], model_tag=data["model_tag"], tags=data["tags"], type=data["type"])

    def to_dict(self) -> Dict[str, Any]:
        return {"name": self.name, "model_tag": self.model_tag, "tags": self.tags, "type": self.type}


@dataclass
class StimulusStatsQuestion:
    title: str
    order: int

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "StimulusStatsQuestion":
        required_keys = ["title", "order"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Invalid data format for StimulusStatsQuestion: {data}")
        return StimulusStatsQuestion(title=data["title"], order=data["order"])

    def to_dict(self) -> Dict[str, Any]:
        return {"title": self.title, "order": self.order}


@dataclass
class StimulusStats:
    files: List[StimulusStatsFile]
    question: StimulusStatsQuestion
    mean: Optional[float] = None
    median: Optional[float] = None
    std: Optional[float] = None
    sem: Optional[float] = None
    ci_95: Optional[float] = None
    options: Dict[str, Any] = field(default_factory=dict)

    @staticmethod
    @validate_args(data=Rules.dict_not_none)
    def from_dict(data: Dict[str, Any]) -> "StimulusStats":
        required_keys = ["files", "question"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Invalid data format for StimulusStats: {data}")

        files = [StimulusStatsFile.from_dict(file) for file in data["files"]]
        question = StimulusStatsQuestion.from_dict(data["question"])
        stats = StimulusStats(
            files=files,
            question=question,
        )

        optional_fields = ["mean", "median", "std", "sem", "ci_95"]
        for field in optional_fields:
            if field in data:
                setattr(stats, field, data[field])

        extra_fields = {k: v for k, v in data.items() if k not in required_keys + optional_fields}
        if extra_fields:
            stats.options.update(extra_fields)

        return stats

    def to_dict(self) -> Dict[str, Any]:
        result = {
            "question": self.question.to_dict(),
            "files": [file.to_dict() for file in self.files],
        }

        optional_fields = ["mean", "median", "std", "sem", "ci_95"]
        for field in optional_fields:
            value = getattr(self, field)
            if value is not None:
                result[field] = value

        if self.options:
            result.update({"options": self.options})

        return result



from typing import List

from requests import HTTPError

from podonos.core.api import APIClient
from podonos.core.base import log
from podonos.core.script import ScriptCreateRequestDto
from podonos.entity.script import ScriptEntity
from podonos.common.validator import Rules, validate_args


class ScriptService:
    def __init__(self, api_client: APIClient):
        self.api_client = api_client

    @validate_args(collection_id=Rules.uuid_not_none, texts=Rules.list_not_none)
    def create_all(self, collection_id: str, texts: List[str]) -> List[ScriptEntity]:
        """
        Create new scripts for a collection

        Args:
            collection_id: The ID of the collection
            texts: List of texts to create scripts for
                - The length of text must be over 1 character

        Returns:
            List of created scripts

        Raises:
            HTTPError: If the request fails
        """
        log.info(f"Create scripts for collection: {collection_id}")

        try:
            request = ScriptCreateRequestDto.from_dict(collection_id, texts)
            response = self.api_client.post("scripts", data=request.to_create_request_dto())
            response.raise_for_status()
            return [ScriptEntity.from_dict(item) for item in response.json()]
        except Exception as e:
            raise HTTPError(f"Failed to create scripts: {e}")

    @validate_args(collection_id=Rules.uuid_not_none)
    def list(self, collection_id: str) -> List[ScriptEntity]:
        """
        Get the list of all scripts

        Returns:
            List of all scripts

        Raises:
            HTTPError: If the request fails
        """
        log.debug(f"Get the list of scripts for collection: {collection_id}")
        try:
            response = self.api_client.get(f"scripts?collection-id={collection_id}")
            response.raise_for_status()
            return [ScriptEntity.from_dict(item) for item in response.json()]
        except Exception as e:
            raise HTTPError(f"Failed to get the list of scripts: {e}")



from typing import Any, Dict, Literal, Optional, List, Union

from podonos.common.validator import Rules, validate_args
from podonos.core.api import APIClient
from podonos.core.base import *
from podonos.core.config import EvalConfigDefault
from podonos.core.evaluator import Evaluator
from podonos.evaluation import AIEvaluation, HumanEvaluation
from podonos.service import CollectionService, EvaluationService, ScriptService, TemplateService


class Client:
    """Podonos Client class. Used for creating individual evaluator and managing the evaluations."""

    _api_client: APIClient
    _initialized: bool = False

    # Services
    _collection_service: CollectionService
    _evaluation_service: EvaluationService
    _script_service: ScriptService
    _template_service: TemplateService

    # Evaluators
    _ai_evaluation: AIEvaluation
    _human_evaluation: HumanEvaluation

    def __init__(self, api_client: APIClient):
        self._api_client = api_client
        self._initialized = True
        self._collection_service = CollectionService(self._api_client)
        self._evaluation_service = EvaluationService(self._api_client)
        self._script_service = ScriptService(self._api_client)
        self._template_service = TemplateService(self._api_client)

        self._ai_evaluation = AIEvaluation(self._api_client)
        self._human_evaluation = HumanEvaluation(self._api_client, self._evaluation_service, self._template_service)

    @validate_args(
        name=Rules.str_not_none_or_none,
        desc=Rules.str_not_none_or_none,
        type=Rules.str_non_empty,
        lan=Rules.str_non_empty,
        granularity=Rules.float_not_none,
        num_eval=Rules.positive_not_none,
        due_hours=Rules.positive_not_none,
        use_annotation=Rules.bool_not_none,
        use_loudness_normalization=Rules.bool_not_none,
        auto_start=Rules.bool_not_none,
        max_upload_workers=Rules.positive_not_none,
    )
    def create_evaluator(
        self,
        name: Optional[str] = None,
        desc: Optional[str] = None,
        type: str = EvalConfigDefault.TYPE.value,
        lan: str = EvalConfigDefault.LAN.value,
        granularity: float = EvalConfigDefault.GRANULARITY,
        num_eval: int = EvalConfigDefault.NUM_EVAL,
        due_hours: int = EvalConfigDefault.DUE_HOURS,
        use_annotation: bool = EvalConfigDefault.USE_ANNOTATION,
        use_loudness_normalization: bool = EvalConfigDefault.USE_LOUDNESS_NORMALIZATION,
        auto_start: bool = EvalConfigDefault.AUTO_START,
        max_upload_workers: int = EvalConfigDefault.MAX_UPLOAD_WORKERS,
    ) -> Evaluator:
        """Creates a new evaluator with a unique evaluation session ID.
        For the language code, see https://www.podonos.com/docs/reference#param-lan

        Args:
            name: This session name. Its length must be > 1. If empty, a random name is used. Optional.
            desc: Description of this session. Optional.
            type: Evaluation type. Default: NMOS
            lan: Human language for this audio. One of those in Language. Default: en-us
            granularity: Granularity of the evaluation scales. Either {1, 0.5}
            num_eval: The minimum number of repetition for each audio evaluation. Should be >=1. Default: 10.
            due_hours: An expected number of days of finishing this mission and getting the evaluation report.
                        Must be >= 12. Default: 12.
            use_annotation: Enable detailed annotation on script for detailed comments.
            use_loudness_normalization: Enable loudness normalization for evaluation.
            auto_start: The evaluation start automatically if True. Otherwise, manually start in the workspace.
            max_upload_workers: The maximum number of upload workers. Must be a positive integer. Default: 20

        Returns:
            Evaluator instance.

        Raises:
            ValueError: if this function is called before calling init().
        """

        if not self._initialized:
            raise ValueError("This function is called before initialization.")
        return self._human_evaluation.create(
            name,
            desc,
            type,
            lan,
            granularity,
            num_eval,
            due_hours,
            use_annotation,
            use_loudness_normalization,
            auto_start,
            max_upload_workers,
        )

    @validate_args(
        name=Rules.str_non_empty,
        template_id=Rules.str_non_empty,
        num_eval=Rules.positive_not_none,
        desc=Rules.str_not_none_or_none,
        auto_start=Rules.bool_not_none,
        max_upload_workers=Rules.positive_not_none,
        use_annotation=Rules.bool_not_none,
        use_loudness_normalization=Rules.bool_not_none,
    )
    def create_evaluator_from_template(
        self,
        name: str,
        template_id: str,
        num_eval: int = EvalConfigDefault.NUM_EVAL,
        desc: Optional[str] = None,
        auto_start: bool = EvalConfigDefault.AUTO_START,
        max_upload_workers: int = EvalConfigDefault.MAX_UPLOAD_WORKERS,
        use_annotation: bool = EvalConfigDefault.USE_ANNOTATION,
        use_loudness_normalization: bool = EvalConfigDefault.USE_LOUDNESS_NORMALIZATION,
    ) -> Evaluator:
        """
        Creates a new evaluator using a predefined template.

        Args:
            name: This session name. Required.
            desc: Description of this session. Optional.
            template_id: The ID of the template to use for evaluation parameters.
            num_eval: The number of evaluators per file. Should be >= 1. Default: 10
            auto_start: The evaluation start automatically if True. Otherwise, manually start in the workspace.
            max_upload_workers: The maximum number of upload workers. Must be a positive integer. Default: 20
            use_annotation: Enable detailed annotation on script for detailed comments. Default: False
            use_loudness_normalization: Enable loudness normalization for evaluation. Default: True

        Returns:
            Evaluator instance.

        Raises:
            ValueError: If the template ID is invalid or not found.
        """
        if not self._initialized:
            raise ValueError("This function is called before initialization.")

        return self._human_evaluation.create_from_template(
            name, template_id, num_eval, desc, use_annotation, use_loudness_normalization, auto_start, max_upload_workers
        )

    @validate_args(
        json=Rules.dict_not_none_or_none,
        json_file=Rules.str_not_none_or_none,
        name=Rules.str_not_none_or_none,
        custom_type=Rules.str_not_none,
        desc=Rules.str_not_none_or_none,
        lan=Rules.str_non_empty,
        num_eval=Rules.positive_not_none,
        use_annotation=Rules.bool_not_none,
        use_loudness_normalization=Rules.bool_not_none,
        auto_start=Rules.bool_not_none,
        max_upload_workers=Rules.positive_not_none,
    )
    def create_evaluator_from_template_json(
        self,
        json: Optional[Dict[str, Any]] = None,
        json_file: Optional[str] = None,
        name: Optional[str] = None,
        custom_type: Union[Literal["SINGLE"], Literal["DOUBLE"]] = "SINGLE",
        desc: Optional[str] = None,
        lan: str = EvalConfigDefault.LAN.value,
        num_eval: int = EvalConfigDefault.NUM_EVAL,
        use_annotation: bool = EvalConfigDefault.USE_ANNOTATION,
        use_loudness_normalization: bool = EvalConfigDefault.USE_LOUDNESS_NORMALIZATION,
        auto_start: bool = EvalConfigDefault.AUTO_START,
        max_upload_workers: int = EvalConfigDefault.MAX_UPLOAD_WORKERS,
    ) -> Evaluator:
        """Creates a new evaluator using a template JSON.

        Args:
            json: Template JSON as a dictionary. Optional if json_file is provided.
            json_file: Path to the JSON template file. Optional if json is provided.
            name: This evaluation name. Required.
            custom_type: Type of evaluation ("SINGLE" or "DOUBLE")
            desc: Description of this evaluation. Optional.
            lan: Language for evaluation. Defaults to EvalConfigDefault.LAN.value.
            num_eval: The number of evaluators per file. Should be >=1.
            use_annotation: Enable detailed annotation on script for detailed rating reasoning.
            use_loudness_normalization: Enable loudness normalization for evaluation. Default: False
            auto_start: The evaluation start automatically if True. Otherwise, manually start in the workspace.
            max_upload_workers: The maximum number of upload workers. Must be a positive integer. Default: 20

        Returns:
            Evaluator instance.

        Raises:
            ValueError: If neither json nor json_file is provided, or if both are provided
            ValueError: If custom_type is not "SINGLE" or "DOUBLE"
            ValueError: If the JSON is invalid or contains incompatible question types
            FileNotFoundError: If the json_file path doesn't exist
        """
        if not self._initialized:
            raise ValueError("This function is called before initialization.")

        return self._human_evaluation.create_from_template_json(
            json=json,
            json_file=json_file,
            name=name,
            custom_type=custom_type,
            desc=desc,
            lan=lan,
            num_eval=num_eval,
            use_annotation=use_annotation,
            use_loudness_normalization=use_loudness_normalization,
            auto_start=auto_start,
            max_upload_workers=max_upload_workers,
        )

    def get_evaluation_list(self) -> List[Dict[str, Any]]:
        """Gets a list of evaluations.

        Args: None

        Returns:
            Evaluation containing all the evaluation info
        """
        return self._evaluation_service.get_evaluation_list()

    @validate_args(evaluation_id=Rules.uuid_not_none, group_by=Rules.str_non_empty)
    def get_stats_json_by_id(self, evaluation_id: str, group_by: Literal["question", "script", "model"] = "question") -> List[Dict[str, Any]]:
        """Gets a list of evaluation statistics referenced by id.

        Args:
            evaluation_id: Evaluation id. See get_evaluation_list() above.
            group_by: Group by question or script. Default: "question".
                      "script" and "model" are only available for single-question evaluation.

        Returns:
            List of statistics for the evaluation.
        """
        return self._evaluation_service.get_stats_json_by_id(evaluation_id, group_by)

    @validate_args(evaluation_id=Rules.uuid_not_none, output_dir=Rules.str_not_none)
    def download_evaluation_files_by_evaluation_id(self, evaluation_id: str, output_dir: str) -> str:
        """Download evaluation files"""
        return self._evaluation_service.download_evaluation_files_by_evaluation_id(evaluation_id, output_dir)

    @validate_args(template_id=Rules.str_non_empty)
    def get_eval_template_info(self, template_id: str) -> Dict[str, Any]:
        """Gets detailed information on the evaluation template by id.

        Args:
            template_id: Evaluation template ID.

        Returns:
            JSON containing the evaluation template info.

        Raises:


        """
        try:
            template = self._template_service.get_template_by_code(template_id)
        except:
            raise ValueError(f"Cannot find the template. Please check the id {template_id}.")
        json = {
            "id": template.id,
            "code": template.code,
            "title": template.title,
            "description": template.description,
            "language": template.language,
            "created_time": template.created_time,
            "updated_time": template.updated_time,
        }
        if template.batch_size == 1:
            json["eval_type"] = "Single"
        elif template.batch_size == 2:
            json["eval_type"] = "Double"
        elif template.batch_size == 3:
            json["eval_type"] = "Triple"
        else:
            ValueError(f"Unknown eval type (batch_size): {template.batch_size}.")

        return json



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Literal, Optional, List, Dict, Any

from podonos.common.enum import QuestionResponseCategory, QuestionUsageType, InstructionCategory, QuestionRelatedModel
from podonos.core.types import QuestionMetadataColumn, QuestionMetadataLinearScale, QuestionMetadataPosition, TemplateQuestion, TemplateOption
from podonos.common.validator import Rules, validate_args

TYPE_OF_OPTION_KEY = Literal["score", "label_text", "reference_file"]
TYPE_OF_QUESTION_KEY = Literal[
    "type",
    "question",
    "instruction",
    "description",
    "scale",
    "allow_multiple",
    "has_other",
    "has_none",
    "related_model",
    "order",
    "options",
    "reference_file",
    "reference_files",
    "anchor_label",
]
TYPE_OF_REFERENCE_FILES = Optional[List[Dict[Literal["path", "type"], str]]]


@dataclass
class Option:
    value: str
    label_text: Optional[str] = None
    order: int = 0
    reference_file: Optional[str] = None

    @classmethod
    def from_dict(cls, data: Dict[TYPE_OF_OPTION_KEY, Any], value: Optional[str] = None, order: int = 0, is_score: bool = True) -> "Option":
        if not data.get("label_text"):
            raise ValueError("Option must have a non-empty 'label_text' field")
        if not is_score:
            return cls(value=data["label_text"], order=order, reference_file=data.get("reference_file", None))
        if not value:
            raise ValueError("Score question's option must have a value")
        return cls(value=value, label_text=data.get("label_text"), order=order, reference_file=data.get("reference_file", None))


class Question(ABC):
    def __init__(self, title: str, type: str, batch_size: int, description: Optional[str] = None, order: int = 0):
        self.title = title  # title can be 'question' or 'instruction'
        self.type = type
        self.batch_size = batch_size
        self.description = description
        self.order = order

    @abstractmethod
    def validate(self) -> None:
        """Validate the question details."""
        if not self.title:
            raise ValueError("Question must have a 'question' or 'instruction'")

    @abstractmethod
    def to_template_question(self) -> TemplateQuestion:
        """Convert to TemplateQuestion."""
        pass

    @classmethod
    @validate_args(data=Rules.dict_not_none, batch_size=Rules.positive_not_none)
    def from_dict(cls, data: Dict[TYPE_OF_QUESTION_KEY, Any], batch_size: int) -> "Question":
        """Create appropriate Question instance from dictionary."""
        question_type = data.get("type")
        if not question_type:
            raise ValueError("Question must have a type")

        question_map = {
            "SCORED": ScoredQuestion,
            "NON_SCORED": NonScoredQuestion,
            "COMPARISON": ComparisonQuestion,
            InstructionCategory.DO.value: Instruction,
            InstructionCategory.WARNING.value: Instruction,
            InstructionCategory.DONT.value: Instruction,
            InstructionCategory.EXAMPLE.value: Instruction,
        }

        if question_type not in question_map:
            raise ValueError(f"Unknown question type: {question_type} (must be one of {', '.join(question_map.keys())})")

        return question_map[question_type].from_dict(data, batch_size)


class ScoredQuestion(Question):
    def __init__(
        self,
        question: str,
        options: List[Option],
        related_model: QuestionRelatedModel,
        batch_size: int,
        description: Optional[str] = None,
        order: int = 0,
    ):
        super().__init__(question, "SCORED", batch_size, description, order)
        self.options = options
        self.related_model = related_model

    def validate(self) -> None:
        super().validate()
        if not self.options:
            raise ValueError("SCORED question must have options")

        if not QuestionRelatedModel.is_member(self.related_model.value) and self.batch_size > 1:
            raise ValueError(
                f"SCORED question must have one of the following valid related_models: {', '.join([item.value for item in QuestionRelatedModel])}"
            )

        # Validate that all option values are numbers
        for option in self.options:
            try:
                float(option.value)
            except ValueError:
                raise ValueError(f"SCORED question option value '{option.value}' must be a number")

    def to_template_question(self) -> TemplateQuestion:
        return TemplateQuestion(
            title=self.title,  # question is stored as 'title' in database
            description=self.description,
            response_category=QuestionResponseCategory.CHOICE_ONE,
            usage_type=QuestionUsageType.SCORE,
            order=self.order,
            related_model=self.related_model,
            options=[
                TemplateOption(value=opt.value, label_text=opt.label_text, order=i, reference_file=opt.reference_file)
                for i, opt in enumerate(self.options)
            ],
        )

    @classmethod
    @validate_args(data=Rules.dict_not_none, batch_size=Rules.positive_not_none)
    def from_dict(cls, data: Dict[TYPE_OF_QUESTION_KEY, Any], batch_size: int) -> "ScoredQuestion":
        if "options" not in data or not data["options"]:
            raise ValueError("SCORED question must have options")

        option_length = len(data["options"])
        if option_length < 1 or option_length > 9:
            raise ValueError("SCORED question must have between 1 and 9 options")

        related_model = (
            QuestionRelatedModel.from_value(data.get("related_model", QuestionRelatedModel.ALL.value)) if batch_size > 1 else QuestionRelatedModel.ALL
        )
        option_values = [str(score) for score in range(option_length, 0, -1)]
        options = [Option.from_dict(opt, value=option_values[i], order=i, is_score=True) for i, opt in enumerate(data["options"])]
        return cls(
            question=data["question"],
            description=data.get("description"),
            options=options,
            related_model=related_model,
            batch_size=batch_size,
            order=data.get("order", 0),
        )


class NonScoredQuestion(Question):
    def __init__(
        self,
        question: str,
        options: List[Option],
        allow_multiple: bool,
        related_model: QuestionRelatedModel,
        batch_size: int,
        description: Optional[str] = None,
        has_other: bool = False,
        has_none: bool = False,
        order: int = 0,
    ):
        super().__init__(question, "NON_SCORED", batch_size, description, order)
        self.options = options
        self.allow_multiple = allow_multiple
        self.has_other = has_other
        self.has_none = has_none
        self.related_model = related_model

    def validate(self) -> None:
        super().validate()
        if not self.options:
            raise ValueError("NON_SCORED question must have options")

        if not QuestionRelatedModel.is_member(self.related_model.value) and self.batch_size > 1:
            raise ValueError(
                f"NON_SCORED question must have one of the following valid related_models: {', '.join([item.value for item in QuestionRelatedModel])}"
            )

    def to_template_question(self) -> TemplateQuestion:
        response_category = QuestionResponseCategory.CHOICE_MULTI if self.allow_multiple else QuestionResponseCategory.CHOICE_ONE_NO_SCORE
        return TemplateQuestion(
            title=self.title,  # question is stored as 'title' in database
            description=self.description,
            response_category=response_category,
            usage_type=QuestionUsageType.SCORE,
            order=self.order,
            related_model=self.related_model,
            has_other=self.has_other,
            has_none=self.has_none,
            options=[
                TemplateOption(value=opt.value, label_text=opt.label_text, order=i, reference_file=opt.reference_file)
                for i, opt in enumerate(self.options)
            ],
        )

    @classmethod
    @validate_args(data=Rules.dict_not_none, batch_size=Rules.positive_not_none)
    def from_dict(cls, data: Dict[TYPE_OF_QUESTION_KEY, Any], batch_size: int) -> "NonScoredQuestion":
        if "options" not in data or not data["options"]:
            raise ValueError("NON_SCORED question must have options")
        if "allow_multiple" not in data:
            raise ValueError("NON_SCORED question must specify allow_multiple")

        option_length = len(data["options"])
        if option_length < 1 or option_length > 9:
            raise ValueError("NON_SCORED question must have between 1 and 9 options")

        related_model = (
            QuestionRelatedModel.from_value(data.get("related_model", QuestionRelatedModel.ALL.value)) if batch_size > 1 else QuestionRelatedModel.ALL
        )
        options = [Option.from_dict(opt, value=None, order=i, is_score=False) for i, opt in enumerate(data["options"])]
        return cls(
            question=data["question"],
            description=data.get("description"),
            options=options,
            allow_multiple=data["allow_multiple"],
            related_model=related_model,
            batch_size=batch_size,
            has_other=data.get("has_other", False),
            has_none=data.get("has_none", False),
            order=data.get("order", 0),
        )


class ComparisonQuestion(Question):
    def __init__(
        self,
        question: str,
        meta_data: QuestionMetadataColumn,
        related_model: QuestionRelatedModel,
        batch_size: int,
        scale: int = 5,
        description: Optional[str] = None,
        order: int = 0,
    ):
        super().__init__(question, "COMPARISON", batch_size, description, order)
        self.scale = scale
        self.meta_data = meta_data
        self.related_model = related_model

    def validate(self) -> None:
        super().validate()
        if self.scale < 2 or self.scale > 9:
            raise ValueError("COMPARISON question scale must be between 2 and 9")

        if not QuestionRelatedModel.is_member(self.related_model.value) and self.batch_size > 1:
            raise ValueError(
                f"COMPARISON question must have one of the following valid related_models: {', '.join([item.value for item in QuestionRelatedModel])}"
            )

    def to_template_question(self) -> TemplateQuestion:
        return TemplateQuestion(
            title=self.title,  # question is stored as 'title' in database
            description=self.description,
            response_category=QuestionResponseCategory.SCALE_LINEAR,
            usage_type=QuestionUsageType.SCORE,
            order=self.order,
            scale=self.scale,
            meta_data=self.meta_data,
            related_model=self.related_model,
        )

    @classmethod
    @validate_args(data=Rules.dict_not_none, batch_size=Rules.positive_not_none)
    def from_dict(cls, data: Dict[TYPE_OF_QUESTION_KEY, Any], batch_size: int) -> "ComparisonQuestion":
        error_message = "COMPARISON question must have 'anchor_label' in the format: {'anchor_label': {'title': optional string, 'label_text': {'left': string, 'right': string}}}"
        if "anchor_label" not in data or "label_text" not in data["anchor_label"]:
            raise ValueError(error_message)

        label_text = data["anchor_label"]["label_text"]
        if "left" not in label_text or "right" not in label_text:
            raise ValueError(error_message)

        related_model = (
            QuestionRelatedModel.from_value(data.get("related_model", QuestionRelatedModel.ALL.value)) if batch_size > 1 else QuestionRelatedModel.ALL
        )
        return cls(
            question=data["question"],
            description=data.get("description"),
            related_model=related_model,
            batch_size=batch_size,
            scale=data.get("scale", 5),
            meta_data=QuestionMetadataColumn(
                linear_scale=QuestionMetadataLinearScale(
                    title=data.get("anchor_label", {}).get("title", None),
                    label_text=QuestionMetadataPosition(left=f"A {label_text['left']}", right=f"B {label_text['right']}"),
                )
            ),
            order=data.get("order", 0),
        )


class Instruction(Question):
    def __init__(
        self,
        instruction: str,
        category: InstructionCategory,
        batch_size: int,
        description: Optional[str] = None,
        order: int = 0,
        reference_files: TYPE_OF_REFERENCE_FILES = None,
    ):
        super().__init__(instruction, "INSTRUCTION", batch_size, description, order)
        self.category = category
        self.reference_files = reference_files

    def validate(self) -> None:
        super().validate()
        if not isinstance(self.category, InstructionCategory):  # type: ignore
            raise ValueError(
                f"Invalid instruction type: {self.category}: Use one of {', '.join([InstructionCategory.DO.value, InstructionCategory.WARNING.value, InstructionCategory.DONT.value])}"
            )

    def to_template_question(self) -> TemplateQuestion:
        # Map GuideCategory to QuestionUsageType
        usage_type_map: Dict[InstructionCategory, QuestionUsageType] = {
            InstructionCategory.DO: QuestionUsageType.GUIDELINE_CORRECT,
            InstructionCategory.WARNING: QuestionUsageType.GUIDELINE_WARNING,
            InstructionCategory.DONT: QuestionUsageType.GUIDELINE_PROHIBIT,
            InstructionCategory.EXAMPLE: QuestionUsageType.GUIDELINE_EXAMPLE,
        }

        return TemplateQuestion(
            title=self.title,  # instruction is stored as 'title' in database
            description=self.description,
            response_category=QuestionResponseCategory.INSTRUCTION,
            usage_type=usage_type_map[self.category],
            order=self.order,
            reference_files=self.reference_files,
            related_model=None,  # Instruction is not related to any model
        )

    @classmethod
    @validate_args(data=Rules.dict_not_none, batch_size=Rules.positive_not_none)
    def from_dict(cls, data: Dict[TYPE_OF_QUESTION_KEY, Any], batch_size: int) -> "Instruction":
        try:
            category = InstructionCategory(data["type"])
        except ValueError:
            raise ValueError(
                f"Invalid instruction type: {data['type']}: Use one of {', '.join([InstructionCategory.DO.value, InstructionCategory.WARNING.value, InstructionCategory.DONT.value])}"
            )

        if "reference_file" in data:
            raise ValueError(f"The 'reference_file' field is not allowed for instruction questions. Please use 'reference_files' instead.")

        if "reference_files" in data and not isinstance(data["reference_files"], list):
            raise ValueError("Reference files must be a List")

        for reference_file in data.get("reference_files", []):
            if not isinstance(reference_file, Dict):
                raise ValueError("Reference files must be a List of Dict")
            if "path" not in reference_file or "type" not in reference_file:
                raise ValueError("Reference files must have 'path' and 'type' fields")
            if reference_file["type"] not in ["reference", "target", "audio"]:
                raise ValueError("Reference file type must be one of the following: reference, target, audio")

        if len(data.get("reference_files", [])) > 3:
            raise ValueError("Instruction questions can have at most 3 reference files")

        return cls(
            instruction=data["instruction"],
            description=data.get("description"),
            category=category,
            batch_size=batch_size,
            order=data.get("order", 0),
            reference_files=data.get("reference_files", None),
        )



"""
Default enum values across whole SDK
"""

from enum import Enum
from typing import List


class EvalType(Enum):
    # Human evaluation types
    NMOS = "NMOS"
    QMOS = "QMOS"
    P808 = "P808"
    SMOS = "SMOS"
    PREF = "PREF"
    CMOS = "CMOS"
    DMOS = "DMOS"
    CSMOS = "CSMOS"
    CUSTOM_SINGLE = "CUSTOM_SINGLE"
    CUSTOM_DOUBLE = "CUSTOM_DOUBLE"

    # AI evaluation types
    ASR = "ASR"

    def get_type(self) -> str:
        if self.value in ["CUSTOM_SINGLE", "CUSTOM_DOUBLE"]:
            return "CUSTOM"
        elif self.value == "PREF":
            return "SPEECH_PREFERENCE"
        return f"SPEECH_{self.value}"

    @staticmethod
    def get_type_by_batch_size(batch_size: int) -> "EvalType":
        if batch_size == 1:
            return EvalType.CUSTOM_SINGLE
        elif batch_size == 2:
            return EvalType.CUSTOM_DOUBLE
        elif batch_size == 3:
            return EvalType.CSMOS
        else:
            raise ValueError(f"Invalid batch size: {batch_size}. Use one of the following: {', '.join([item.value for item in EvalType])}")

    @staticmethod
    def get_single_types() -> List["EvalType"]:
        """Get all single stimulus evaluation types"""
        return [EvalType.NMOS, EvalType.QMOS, EvalType.P808, EvalType.CUSTOM_SINGLE]

    @staticmethod
    def get_double_types() -> List["EvalType"]:
        """Get all double stimuli evaluation types"""
        return [EvalType.PREF, EvalType.SMOS, EvalType.CMOS, EvalType.CUSTOM_DOUBLE]

    @staticmethod
    def get_triple_types() -> List["EvalType"]:
        """Get all triple stimuli evaluation types"""
        return [EvalType.CSMOS]

    @staticmethod
    def is_single(type_str: str) -> bool:
        """Check if type is single stimulus"""
        return EvalType(type_str) in EvalType.get_single_types()

    @staticmethod
    def is_double(type_str: str) -> bool:
        """Check if type is double stimuli"""
        return EvalType(type_str) in EvalType.get_double_types()

    @staticmethod
    def is_triple(type_str: str) -> bool:
        """Check if type is triple stimuli"""
        return EvalType(type_str) in EvalType.get_triple_types()

    @staticmethod
    def is_eval_type(eval_type: str) -> bool:
        return any([item for item in EvalType if item.value == eval_type])


class AIEvalType(Enum):
    ASR = "ASR"
    ALL = "ALL"

    @classmethod
    def from_value(cls, value: str) -> "AIEvalType":
        for member in cls:
            if member.value == value:
                return member
        raise ValueError(f"{value} is not a valid value for {cls.__name__}")

    @staticmethod
    def is_ai_type(eval_type: "AIEvalType") -> bool:
        return any([item for item in AIEvalType if item == eval_type])


class Language(Enum):
    ENGLISH_AMERICAN = "en-us"
    ENGLISH_BRITISH = "en-gb"
    ENGLISH_AUSTRALIAN = "en-au"
    ENGLISH_CANADIAN = "en-ca"
    ENGLISH_INDIA = "en-in"
    PORTUGUESE_PORTUGAL = "pt-pt"
    PORTUGUESE_BRAZIL = "pt-br"
    KOREAN = "ko-kr"
    MANDARIN = "zh-cn"
    SPANISH_SPAIN = "es-es"
    SPANISH_MEXICO = "es-mx"
    FRENCH = "fr-fr"
    FRENCH_CANADA = "fr-ca"
    GERMAN = "de-de"
    JAPANESE = "ja-jp"
    ITALIAN = "it-it"
    POLISH = "pl-pl"
    AUDIO = "audio"

    @classmethod
    def from_value(cls, value: str) -> "Language":
        for member in cls:
            if member.value == value:
                return member
        raise ValueError(f"{value} is not a valid value for {cls.__name__}")

    @staticmethod
    def values() -> List[str]:
        return [item.value for item in Language]


class QuestionFileType(str, Enum):
    STIMULUS = "STIMULUS"
    REF = "REF"
    META = "META"


class QuestionResponseCategory(Enum):
    CHOICE_ONE = "CHOICE_ONE"
    CHOICE_MULTI = "CHOICE_MULTI"
    CHOICE_ONE_NO_SCORE = "CHOICE_ONE_NO_SCORE"
    SCALE_LINEAR = "SCALE_LINEAR"
    INSTRUCTION = "INSTRUCTION"


class QuestionUsageType(Enum):
    GUIDELINE_EXAMPLE = "GUIDELINE_EXAMPLE"
    GUIDELINE_CORRECT = "GUIDELINE_CORRECT"
    GUIDELINE_WARNING = "GUIDELINE_WARNING"
    GUIDELINE_PROHIBIT = "GUIDELINE_PROHIBIT"
    SCORE = "SCORE"

    @staticmethod
    def is_score(usage_type: "QuestionUsageType") -> bool:
        return usage_type == QuestionUsageType.SCORE


class InstructionCategory(Enum):
    EXAMPLE = "EXAMPLE"
    DO = "DO"
    WARNING = "WARNING"
    DONT = "DONT"


class QuestionRelatedModel(Enum):
    ALL = "ALL"
    MODEL_A = "MODEL_A"
    MODEL_B = "MODEL_B"

    @staticmethod
    def from_value(value: str) -> "QuestionRelatedModel":
        for member in QuestionRelatedModel:
            if member.value == value:
                return member
        raise ValueError(f"Invalid related model: {value}. Use one of the following: {', '.join([item.value for item in QuestionRelatedModel])}")

    @staticmethod
    def is_member(value: str) -> bool:
        return value in [item.value for item in QuestionRelatedModel]


class CollectionTarget(Enum):
    AUDIO = "AUDIO"

    @staticmethod
    def from_value(value: str) -> "CollectionTarget":
        for member in CollectionTarget:
            if member.value == value:
                return member
        raise ValueError(f"Invalid collection target: {value}. Use one of the following: {', '.join([item.value for item in CollectionTarget])}")


class CollectionCustomerStatus(Enum):
    DRAFT = "DRAFT"
    ACTIVE = "ACTIVE"
    COMPLETED = "COMPLETED"
    CANCELLED = "CANCELLED"


class SpeechStyle(str, Enum):
    NONE = "NONE"
    NARRATIVE = "NARRATIVE"
    CONVERSATION = "CONVERSATION"
    POEM = "POEM"
    STORY = "STORY"
    NEWS = "NEWS"
    ADVERTISEMENT = "ADVERTISEMENT"


class SpeechEmotion(str, Enum):
    NONE = "NONE"
    NEUTRAL = "NEUTRAL"
    HAPPY = "HAPPY"
    SAD = "SAD"
    ANGRY = "ANGRY"
    SURPRISED = "SURPRISED"
    DISGUSTED = "DISGUSTED"
    FEARFUL = "FEARFUL"
    EXCITED = "EXCITED"
    SARCASTIC = "SARCASTIC"
    ENCOURAGING = "ENCOURAGING"
    PLAYFUL = "PLAYFUL"
    SERIOUS = "SERIOUS"


class SpeechSpeed(str, Enum):
    SLOW = "SLOW"
    NORMAL = "NORMAL"
    FAST = "FAST"



import json as json_lib
from pathlib import Path
from dataclasses import dataclass
from datetime import datetime
from typing import Literal, Optional, List, Dict, Any, Tuple

from podonos.common.enum import Language
from podonos.common.validator import Rules, validate_args
from podonos.core.base import *
from podonos.core.types import TemplateQuestion
from podonos.core.query import NonScoredQuestion, Question, Instruction, ComparisonQuestion, ScoredQuestion
from podonos.core.query import TYPE_OF_REFERENCE_FILES

TYPE_OF_TEMPLATE_KEY = Literal["questions", "instructions"]


@dataclass
class Template:
    """Template class for handling API responses"""

    id: Optional[str] = None
    code: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    language: Optional[Language] = None
    batch_size: Optional[int] = None
    created_time: Optional[datetime] = None
    updated_time: Optional[datetime] = None

    @staticmethod
    @validate_args(data=Rules.dict_not_none)
    def from_api_response(data: Dict[str, Any]) -> "Template":
        """Create Template instance from API response."""
        required_keys = [
            "id",
            "code",
            "title",
            "description",
            "batch_size",
            "language",
            "created_time",
            "updated_time",
        ]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Missing {key} in {data}")

        return Template(
            id=data["id"],
            code=data["code"],
            title=data["title"],
            description=data["description"],
            batch_size=data["batch_size"],
            language=Language.from_value(data["language"]),
            created_time=datetime.fromisoformat(data["created_time"].replace("Z", "+00:00")),
            updated_time=datetime.fromisoformat(data["updated_time"].replace("Z", "+00:00")),
        )


class TemplateJsonLoader:
    """Loader class for template JSON data"""

    @staticmethod
    @validate_args(json=Rules.dict_not_none_or_none, json_file=Rules.str_not_none_or_none)
    def load_json(json: Optional[Dict[str, Any]] = None, json_file: Optional[str] = None) -> Dict[TYPE_OF_TEMPLATE_KEY, Any]:
        """Load template JSON data from a file"""
        # Validate input parameters
        if json is None and json_file is None:
            raise ValueError("Either 'json' or 'json_file' must be provided")
        if json is not None and json_file is not None:
            raise ValueError("Only one of 'json' or 'json_file' should be provided")

        # Get template data
        if json_file is not None:
            log.info(f"Reading template from file: {json_file}")
            json_path = Path(json_file)
            if not json_path.exists():
                raise FileNotFoundError(f"JSON file not found: {json_file}")

            with open(json_path, "r", encoding="utf-8") as f:
                template_data = json_lib.load(f)

        else:
            log.info("Using provided template JSON")
            assert json is not None
            template_data = json
        return template_data  # type: ignore


class TemplateValidator:
    """Validator class for template JSON data"""

    @staticmethod
    @validate_args(data=Rules.dict_not_none, batch_size=Rules.positive_not_none)
    def validate_and_create_questions(
        data: Dict[TYPE_OF_TEMPLATE_KEY, Any], batch_size: int
    ) -> Tuple[List[TemplateQuestion], List[TemplateQuestion]]:
        """Validates the template JSON data and returns TemplateQuestion objects.

        Args:
            data: Template JSON data
            batch_size: Number of stimuli to compare (1 for single, 2 for double, etc.)

        Returns:
            Tuple of (instructions, questions)

        Raises:
            ValueError: If template structure is invalid or contains incompatible questions
        """
        # Validate core questions (required)
        if "questions" not in data or not isinstance(data["questions"], list):
            raise ValueError("Template must contain a 'questions' list")

        question_length = len(data["questions"])  # type: ignore
        if question_length < 1 or question_length > 9:
            raise ValueError("Template must contain between 1 and 9 questions")

        instructions = TemplateValidator.process_questions(data, "instructions", [Instruction], batch_size)
        core_questions = TemplateValidator.process_questions(data, "questions", [ScoredQuestion, NonScoredQuestion, ComparisonQuestion], batch_size)

        log.debug(f"Processed {len(instructions)} instructions and {len(core_questions)} questions")

        return instructions, core_questions

    @staticmethod
    @validate_args(data=Rules.dict_not_none, key=Rules.str_not_none, expected_types=Rules.list_not_none, batch_size=Rules.positive_not_none)
    def process_questions(data: Dict[TYPE_OF_TEMPLATE_KEY, Any], key: str, expected_types: List[type], batch_size: int) -> List[TemplateQuestion]:
        """
        Process questions from the given data dictionary.

        Args:
            data: The data dictionary containing questions.
            key: The key in the data dictionary to process.
            expected_types: The expected types of questions.
            order_start: The starting order number for questions.
            batch_size: The batch size for validation, if applicable.

        Returns:
            A list of processed TemplateQuestion objects.

        Raises:
            ValueError: If the questions are not in the expected format or type.
        """
        if key not in data or not data[key]:
            return []

        if not isinstance(data[key], list):
            raise ValueError(f"{key.capitalize()} must be in a list format")

        log.debug(f"Processing {len(data[key])} {key}...")
        questions: List[TemplateQuestion] = []
        for i, q_data in enumerate(data[key]):
            try:
                question = Question.from_dict(q_data, batch_size)
                question.validate()

                if not any(isinstance(question, expected_type) for expected_type in expected_types):
                    types = ", ".join([expected_type.__name__ for expected_type in expected_types])
                    raise ValueError(f"Question in {key} section must be one of the following types: {types}, got {q_data.get('type')}")

                if isinstance(question, ComparisonQuestion) and batch_size == 1:
                    raise ValueError(
                        "COMPARISON type questions are not allowed in single stimulus evaluation. "
                        "Please use batch_size=2 for comparison questions."
                    )

                if isinstance(question, Instruction) and question.reference_files:
                    TemplateValidator.check_if_reference_file_is_audio_file(None, question.reference_files)

                if (isinstance(question, ScoredQuestion) or isinstance(question, NonScoredQuestion)) and question.options:
                    for option in question.options:
                        if option.reference_file:
                            TemplateValidator.check_if_reference_file_is_audio_file(option.reference_file, None)

                template_question = question.to_template_question()
                template_question.order = i
                questions.append(template_question)
            except Exception as e:
                log.error(f"Failed to process {key} question {i}: {str(e)}")
                raise

        return questions

    @staticmethod
    @validate_args(reference_file=Rules.str_not_none_or_none, reference_files=Rules.list_not_none_or_none)
    def check_if_reference_file_is_audio_file(reference_file: Optional[str], reference_files: TYPE_OF_REFERENCE_FILES) -> None:
        """Check if the reference file is valid"""
        if reference_file is None and reference_files is None:
            return

        if reference_file is not None and reference_files is not None:
            raise ValueError("Only one of 'reference_file' or 'reference_files' should be provided. The instruction only supports 'reference_files'")

        if reference_file is not None:
            if not Path(reference_file).exists():
                raise ValueError(f"Reference file not found: {reference_file}")

            if not Path(reference_file).is_file():
                raise ValueError(f"Reference file is not a file: {reference_file}")

            if not Path(reference_file).suffix.lower() in [".wav", ".mp3", ".flac"]:
                raise ValueError(f"Reference file must be an audio file: {reference_file}. Supported formats: .wav, .mp3, .flac")

        if reference_files is not None:
            for file in reference_files:
                path = file["path"]
                type = file["type"]
                if not Path(path).exists():
                    raise ValueError(f"Reference file not found: {path}")

                if not Path(path).is_file():
                    raise ValueError(f"Reference file is not a file: {path}")

                if not Path(path).suffix.lower() in [".wav", ".mp3", ".flac"]:
                    raise ValueError(f"Reference file must be an audio file: {path}. Supported formats: .wav, .mp3, .flac")

                if type not in ["reference", "target", "audio"]:
                    raise ValueError(f"Reference file type must be one of the following: reference, target, audio. Got {type}")



from dataclasses import dataclass
from typing import Any, Dict, List, Optional


@dataclass
class VerificationErrorDetail:
    code: str
    message: str
    expected: Optional[str] = None
    actual: Optional[str] = None

    @classmethod
    def from_dict(
        cls, data: Optional[Dict[str, Any]]
    ) -> Optional["VerificationErrorDetail"]:
        if data is None:
            return None
        return cls(
            code=data.get("code", ""),
            message=data.get("message", ""),
            expected=data.get("expected"),
            actual=data.get("actual"),
        )


@dataclass
class FileVerificationResult:
    uploaded_file_name: str
    verified: bool
    file_meta_id: Optional[str]
    error: Optional[VerificationErrorDetail]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "FileVerificationResult":
        return cls(
            uploaded_file_name=data["uploaded_file_name"],
            verified=data["verified"],
            file_meta_id=data.get("file_meta_id"),
            error=VerificationErrorDetail.from_dict(data.get("error")),
        )


@dataclass
class VerifyFilesResponse:
    all_verified: bool
    verified_count: int
    failed_count: int
    results: List[FileVerificationResult]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "VerifyFilesResponse":
        return cls(
            all_verified=data["all_verified"],
            verified_count=data["verified_count"],
            failed_count=data["failed_count"],
            results=[FileVerificationResult.from_dict(r) for r in data["results"]],
        )


@dataclass
class ProcessFilesResponse:
    processing_count: int
    file_meta_ids: List[str]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ProcessFilesResponse":
        return cls(
            processing_count=data["processing_count"],
            file_meta_ids=data["file_meta_ids"],
        )



from dataclasses import dataclass
from typing import List, Optional


class NotSupportedError(Exception):
    """Exception raised for unsupported operations."""

    def __init__(self, message: str = "This operation is not supported"):
        self.message = message
        super().__init__(self.message)


class InvalidFileError(Exception):
    """Exception raised for invalid file input."""

    def __init__(self, message: str = "This file is invalid"):
        self.message = message
        super().__init__(self.message)


@dataclass
class FileVerificationFailure:
    uploaded_file_name: str
    original_name: str
    error_code: str
    message: str
    expected: Optional[str] = None
    actual: Optional[str] = None


class UploadVerificationError(Exception):
    """Raised when file upload verification fails."""

    def __init__(
        self,
        message: str,
        failures: List[FileVerificationFailure],
        retry_count: int = 0,
        max_retries: int = 3,
    ):
        self.failures = failures
        self.retry_count = retry_count
        self.max_retries = max_retries

        detail_lines = [
            f"\n  - {f.original_name}: {f.error_code} ({f.message})" for f in failures
        ]
        full_message = f"{message}{''.join(detail_lines)}"
        super().__init__(full_message)


class UploadRetryExhaustedError(UploadVerificationError):
    """Raised when max retry attempts exceeded for upload verification."""

    pass



from typing import Dict, Literal, Optional, Union, Any
from requests import HTTPError

from podonos.common.constant import PODONOS_CONTACT_EMAIL
from podonos.common.enum import EvalType
from podonos.core.api import APIClient
from podonos.core.base import log
from podonos.core.config import EvalConfig, EvalConfigDefault
from podonos.core.evaluator import Evaluator
from podonos.core.template import TemplateJsonLoader, TemplateValidator
from podonos.service import EvaluationService, TemplateService
from podonos.common.validator import Rules, validate_args


class HumanEvaluation:
    _api_client: APIClient
    _evaluation_service: EvaluationService
    _template_service: TemplateService

    def __init__(self, api_client: APIClient, evaluation_service: EvaluationService, template_service: TemplateService):
        self._api_client = api_client
        self._evaluation_service = evaluation_service
        self._template_service = template_service

    @validate_args(
        name=Rules.str_not_none_or_none,
        desc=Rules.str_not_none_or_none,
        type=Rules.str_not_none,
        lan=Rules.str_not_none,
        granularity=Rules.float_not_none,
        num_eval=Rules.int_not_none,
        due_hours=Rules.int_not_none,
        use_annotation=Rules.bool_not_none,
        use_loudness_normalization=Rules.bool_not_none,
        auto_start=Rules.bool_not_none,
        max_upload_workers=Rules.int_not_none,
    )
    def create(
        self,
        name: Optional[str] = None,
        desc: Optional[str] = None,
        type: str = EvalConfigDefault.TYPE.value,
        lan: str = EvalConfigDefault.LAN.value,
        granularity: float = EvalConfigDefault.GRANULARITY,
        num_eval: int = EvalConfigDefault.NUM_EVAL,
        due_hours: int = EvalConfigDefault.DUE_HOURS,
        use_annotation: bool = EvalConfigDefault.USE_ANNOTATION,
        use_loudness_normalization: bool = EvalConfigDefault.USE_LOUDNESS_NORMALIZATION,
        auto_start: bool = EvalConfigDefault.AUTO_START,
        max_upload_workers: int = EvalConfigDefault.MAX_UPLOAD_WORKERS,
    ) -> Evaluator:
        """Creates a new evaluator with a unique evaluation session ID.
        For the language code, see https://www.podonos.com/docs/reference#param-lan

        Args:
            name: This session name. Its length must be > 1. If empty, a random name is used. Optional.
            desc: Description of this session. Optional.
            type: Evaluation type. Default: NMOS
            lan: Human language for this audio. One of those in Language. Default: en-us
            granularity: Granularity of the evaluation scales. Either {1, 0.5}
            num_eval: The minimum number of repetition for each audio evaluation. Should be >=1. Default: 10.
            due_hours: An expected number of days of finishing this mission and getting the evaluation report.
                        Must be >= 12. Default: 12.
            use_annotation: Enable detailed annotation on script for detailed rating reasoning.
            use_loudness_normalization: Enable loudness normalization for evaluation.
            auto_start: The evaluation start automatically if True. Otherwise, manually start in the workspace.
            max_upload_workers: The maximum number of upload workers. Must be a positive integer. Default: 20

        Returns:
            Evaluator instance.

        Raises:
            ValueError: if this type is not supported.
        """

        if not EvalType.is_eval_type(type):
            raise ValueError(
                "Not supported evaluation types. Use one of the " "{'NMOS', 'QMOS', 'P808', 'CMOS', 'SMOS', 'PREF', 'CUSTOM_SINGLE', 'CUSTOM_DOUBLE'}"
            )

        eval_config = EvalConfig(
            name=name,
            desc=desc,
            type=type,
            lan=lan,
            granularity=granularity,
            num_eval=num_eval,
            due_hours=due_hours,
            use_annotation=use_annotation,
            use_loudness_normalization=use_loudness_normalization,
            auto_start=auto_start,
            max_upload_workers=max_upload_workers,
        )

        if EvalType.is_double(type):
            supported_types = EvalType.get_double_types()
        elif EvalType.is_single(type):
            supported_types = EvalType.get_single_types()
        elif EvalType.is_triple(type):
            supported_types = EvalType.get_triple_types()
        else:
            raise ValueError(f"Invalid evaluation type: {type}")

        return Evaluator(api_client=self._api_client, eval_config=eval_config, supported_eval_types=supported_types)

    @validate_args(
        name=Rules.str_not_none,
        template_id=Rules.str_non_empty,
        num_eval=Rules.int_not_none,
        desc=Rules.str_not_none_or_none,
        use_annotation=Rules.bool_not_none,
        use_loudness_normalization=Rules.bool_not_none,
        auto_start=Rules.bool_not_none,
        max_upload_workers=Rules.int_not_none,
    )
    def create_from_template(
        self,
        name: str,
        template_id: str,
        num_eval: int,
        desc: Optional[str] = None,
        use_annotation: bool = EvalConfigDefault.USE_ANNOTATION,
        use_loudness_normalization: bool = EvalConfigDefault.USE_LOUDNESS_NORMALIZATION,
        auto_start: bool = EvalConfigDefault.AUTO_START,
        max_upload_workers: int = EvalConfigDefault.MAX_UPLOAD_WORKERS,
    ) -> Evaluator:
        """
        Creates a new evaluator using a predefined template.

        Args:
            name: This session name. Required.
            desc: Description of this session. Optional.
            template_id: The ID of the template to use for evaluation parameters.
            num_eval: The number of evaluators per file. Should be >= 1.
            auto_start: The evaluation start automatically if True. Otherwise, manually start in the workspace.
            max_upload_workers: The maximum number of upload workers. Must be a positive integer. Default: 20

        Returns:
            Evaluator instance.

        Raises:
            ValueError: If the template ID is invalid or not found.
        """
        if not template_id:
            raise ValueError("Template Id should exist")

        template = self._template_service.get_template_by_code(template_id)
        if template.batch_size is None:
            raise ValueError(f"Template with id {template_id} has no batch size")

        eval_config = EvalConfig(
            type=EvalType.get_type_by_batch_size(template.batch_size).value,
            name=name,
            desc=desc,
            num_eval=num_eval,
            use_annotation=use_annotation,
            use_loudness_normalization=use_loudness_normalization,
            auto_start=auto_start,
            template_id=str(template.id),
            max_upload_workers=max_upload_workers,
        )

        if template.batch_size == 1:
            supported_types = EvalType.get_single_types()
        elif template.batch_size == 2:
            # CMOS isn't supported in create_from_template
            supported_types = [EvalType.SMOS, EvalType.PREF, EvalType.CUSTOM_DOUBLE]
        elif template.batch_size == 3:
            supported_types = EvalType.get_triple_types()
        else:
            raise ValueError(f"Template has invalid type so please contact {PODONOS_CONTACT_EMAIL}")
        return Evaluator(api_client=self._api_client, eval_config=eval_config, supported_eval_types=supported_types)

    @validate_args(
        json=Rules.dict_not_none_or_none,
        json_file=Rules.str_not_none_or_none,
        name=Rules.str_not_none_or_none,
        custom_type=Rules.str_not_none,
        desc=Rules.str_not_none_or_none,
        lan=Rules.str_not_none,
        num_eval=Rules.int_not_none,
        use_annotation=Rules.bool_not_none,
        use_loudness_normalization=Rules.bool_not_none,
        auto_start=Rules.bool_not_none,
        max_upload_workers=Rules.int_not_none,
    )
    def create_from_template_json(
        self,
        json: Optional[Dict[str, Any]] = None,
        json_file: Optional[str] = None,
        name: Optional[str] = None,
        custom_type: Union[Literal["SINGLE"], Literal["DOUBLE"]] = "SINGLE",
        desc: Optional[str] = None,
        lan: str = EvalConfigDefault.LAN.value,
        num_eval: int = EvalConfigDefault.NUM_EVAL,
        use_annotation: bool = EvalConfigDefault.USE_ANNOTATION,
        use_loudness_normalization: bool = EvalConfigDefault.USE_LOUDNESS_NORMALIZATION,
        auto_start: bool = EvalConfigDefault.AUTO_START,
        max_upload_workers: int = EvalConfigDefault.MAX_UPLOAD_WORKERS,
    ) -> Evaluator:
        """Creates a new evaluator using a template JSON.

        Args:
            json: Template JSON as a dictionary. Optional if json_file is provided.
            json_file: Path to the JSON template file. Optional if json is provided.
            name: This evaluation name. Required.
            custom_type: Type of evaluation ("SINGLE" or "DOUBLE")
            desc: Description of this evaluation. Optional.
            lan: Language for evaluation. Defaults to EvalConfigDefault.LAN.value.
            num_eval: The number of evaluators per file. Should be >=1.
            use_annotation: Enable detailed annotation on script for detailed rating reasoning.
            use_loudness_normalization: Enable loudness normalization for evaluation. Default: False
            auto_start: The evaluation start automatically if True. Otherwise, manually start in the workspace.
            max_upload_workers: The maximum number of upload workers. Must be a positive integer. Default: 20

        Returns:
            Evaluator instance.

        Raises:
            ValueError: If neither json nor json_file is provided, or if both are provided
            ValueError: If custom_type is not "SINGLE" or "DOUBLE"
            ValueError: If the JSON is invalid or contains incompatible question types
            FileNotFoundError: If the json_file path doesn't exist
        """
        # Validate custom_type
        if custom_type not in ["SINGLE", "DOUBLE"]:
            raise ValueError('custom_type must be either "SINGLE" or "DOUBLE"')

        eval_type = EvalType.CUSTOM_SINGLE if custom_type == "SINGLE" else EvalType.CUSTOM_DOUBLE
        batch_size = 1 if custom_type == "SINGLE" else 2
        # Load template data
        template_data = TemplateJsonLoader.load_json(json, json_file)

        # Use the validator from template.py
        instructions, core_questions = TemplateValidator.validate_and_create_questions(template_data, batch_size)
        log.info("Template JSON is validated.")

        # Create an evaluator
        eval_config = EvalConfig(
            name=name,
            desc=desc,
            type=eval_type.value,
            lan=lan,
            num_eval=num_eval,
            use_annotation=use_annotation,
            use_loudness_normalization=use_loudness_normalization,
            auto_start=auto_start,
            max_upload_workers=max_upload_workers,
        )
        log.info(f"Created evaluation config with type: {eval_type.value}")

        if custom_type == "SINGLE":
            supported_types = EvalType.get_single_types()
        elif custom_type == "DOUBLE":
            # CMOS isn't supported in create_from_template_json
            supported_types = [EvalType.SMOS, EvalType.PREF, EvalType.CUSTOM_DOUBLE]
        else:
            raise ValueError('custom_type must be either "SINGLE" or "DOUBLE"')

        template_service = TemplateService(self._api_client)
        evaluator = Evaluator(api_client=self._api_client, eval_config=eval_config, supported_eval_types=supported_types)
        try:
            if instructions:
                log.debug(f"Creating {len(instructions)} instructions...")
                instructions = template_service.create_template_questions_by_evaluation_id_and_questions(evaluator.get_evaluation_id(), instructions)
                for instruction in instructions:
                    if instruction.id and instruction.reference_files:
                        template_service.upload_reference_files_by_url_and_file_paths(instruction.reference_files, instruction.id)

            if core_questions:
                log.debug(f"Creating {len(core_questions)} core questions...")
                core_questions = template_service.create_template_questions_by_evaluation_id_and_questions(
                    evaluator.get_evaluation_id(), core_questions
                )

            # Create options for questions that have options
            questions_with_options = [q for q in (instructions + core_questions) if q.options]
            if questions_with_options:
                log.debug(f"Creating options for {len(questions_with_options)} questions...")
                for question in questions_with_options:
                    if question.id:
                        options = template_service.create_template_options_by_question_id_and_options(question.id, question.options)
                        for option in options:
                            if option.id and option.reference_file:
                                presigned_url = template_service.get_presigned_url_by_template_option_id(option.id)
                                template_service.upload_reference_file_by_url_and_file_path(presigned_url, option.reference_file, option_id=option.id)

        except Exception as e:
            log.error(f"Failed to create template: {str(e)}")
            raise HTTPError(f"Failed to create template questions: {e}")

        log.info("Template creation completed successfully")
        return evaluator



from dataclasses import dataclass, field
from typing import Optional, Any, Dict

from podonos.common.enum import CollectionTarget, Language


@dataclass
class CollectionCreateRequestDto:
    name: str
    description: Optional[str] = None
    language: str = field(default=Language.ENGLISH_AMERICAN.value)
    num_required_people: int = field(default=10)
    target: CollectionTarget = field(default=CollectionTarget.AUDIO)

    def get_language(self) -> Language:
        return Language(self.language)

    def to_create_request_dto(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "description": self.description,
            "language": self.language,
            "num_required_people": self.num_required_people,
            "target": self.target.value,
        }

    @staticmethod
    def from_dict(
        name: str,
        description: Optional[str] = None,
        language: str = Language.ENGLISH_AMERICAN.value,
        num_required_people: int = 10,
        target: str = CollectionTarget.AUDIO.value,
    ) -> "CollectionCreateRequestDto":
        if len(name) == 0:
            raise ValueError("The name of the collection is required")

        if num_required_people < 1:
            raise ValueError("The number of required people must be greater than 0")

        if language != Language.ENGLISH_AMERICAN.value:
            raise ValueError("The language of the collection must be en-us")

        if target != CollectionTarget.AUDIO.value:
            raise ValueError("The target of the collection must be one of the following: AUDIO")

        return CollectionCreateRequestDto(
            name=name,
            description=description,
            language=Language.from_value(language).value,
            num_required_people=num_required_people,
            target=CollectionTarget.from_value(target),
        )



from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Any

from podonos.common.enum import SpeechStyle, SpeechEmotion, SpeechSpeed


@dataclass
class ScriptEntity:
    id: str
    collection_id: str
    text: str
    estimated_duration: int
    required_count: int
    collected_count: int
    speech_style: SpeechStyle
    emotion: SpeechEmotion
    speed: SpeechSpeed
    meta_data: Dict[str, Any]
    created_time: datetime
    updated_time: datetime

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "ScriptEntity":
        return ScriptEntity(
            id=data["id"],
            collection_id=data["collection_id"],
            text=data["text"],
            estimated_duration=data["estimated_duration"],
            required_count=data["required_count"],
            collected_count=data["collected_count"],
            speech_style=SpeechStyle(data["speech_style"]),
            emotion=SpeechEmotion(data["emotion"]),
            speed=SpeechSpeed(data["speed"]),
            meta_data=data["meta_data"],
            created_time=data["created_time"],
            updated_time=data["updated_time"],
        )



from typing import Any, Dict, List, Literal, Optional
from podonos.core.api import APIClient
from podonos.core.base import log
from podonos.core.template import Template
from podonos.core.types import TemplateOption, TemplateQuestion
from podonos.common.exception import HTTPError
from podonos.core.query import TYPE_OF_REFERENCE_FILES
from podonos.common.validator import Rules, validate_args


class TemplateService:
    """Service class for handling template"""

    def __init__(self, api_client: APIClient):
        self.api_client = api_client

    @validate_args(template_id=Rules.str_not_none)
    def get_template_by_code(self, template_id: str) -> Template:
        """
        Get template information by Id

        Returns: Template
        """
        try:
            response = self.api_client.get(f"templates/one?code={template_id}")
            response.raise_for_status()
            template = Template.from_api_response(response.json())
            log.info(f"Get template by id {template_id}")
            return template
        except Exception as e:
            raise HTTPError(f"Failed to get template by id: {template_id} / {e}")

    @validate_args(evaluation_id=Rules.uuid_not_none, template_questions=Rules.list_not_none)
    def create_template_questions_by_evaluation_id_and_questions(
        self, evaluation_id: str, template_questions: List[TemplateQuestion]
    ) -> List[TemplateQuestion]:
        """
        Create template questions by evaluation id and question list

        Returns: List of TemplateQuestion
        """
        try:
            log.debug(f"Creating {len(template_questions)} template questions for evaluation {evaluation_id}")
            response = self.api_client.put(
                f"template-questions/bulk", data={"evaluation_id": evaluation_id, "questions": [q.to_create_dict() for q in template_questions]}
            )
            response.raise_for_status()

            log.debug(f"Create template questions by evaluation id {evaluation_id}")
            for q_response, question in zip(response.json(), template_questions):
                question.id = q_response["id"]
            return template_questions
        except Exception as e:
            raise HTTPError(f"Failed to create template questions by evaluation id: {evaluation_id} / {e}")

    @validate_args(template_question_id=Rules.uuid_not_none, template_options=Rules.list_not_none)
    def create_template_options_by_question_id_and_options(
        self, template_question_id: str, template_options: List[TemplateOption]
    ) -> List[TemplateOption]:
        """
        Create template options by question id and option list
        """
        try:
            log.debug(f"Creating {len(template_options)} template options for question {template_question_id}")
            response = self.api_client.put(
                f"template-options/bulk", data={"template_question_id": template_question_id, "options": [opt.to_dict() for opt in template_options]}
            )
            response.raise_for_status()
            log.debug(f"Create template options by question id {template_question_id}")
            for opt_response, option in zip(response.json(), template_options):
                option.id = opt_response["id"]
            return template_options
        except Exception as e:
            raise HTTPError(f"Failed to create template options by question id: {template_question_id} / {e}")

    @validate_args(template_option_id=Rules.uuid_not_none)
    def get_presigned_url_by_template_option_id(self, template_option_id: str) -> str:
        """
        Get presigned url by template option id
        """
        try:
            response = self.api_client.post(f"template-options/{template_option_id}/reference-uri", data={})
            response.raise_for_status()
            return response.text.replace('"', "")
        except Exception as e:
            log.error(f"Failed to get presigned URL for template option id: {template_option_id} / {e}")
            raise HTTPError(f"Failed to get presigned URL for template option id: {template_option_id} / {e}")

    @validate_args(url=Rules.str_not_none, file_path=Rules.file_path_not_none, option_id=Rules.str_not_none_or_none)
    def upload_reference_file_by_url_and_file_path(self, url: str, file_path: str, option_id: Optional[str] = None) -> None:
        """
        Upload reference file by url and file path
        """
        try:
            with open(file_path, "rb") as file:
                upload_response = self.api_client.external_put(url, data=file)
                upload_response.raise_for_status()

            log.debug(f"Successfully uploaded reference file {file_path} of Template by url")
        except Exception as e:
            if option_id:
                self.api_client.delete(f"template-options/{option_id}/reference-uri")
                raise HTTPError(f"Failed to upload reference file by option id: {option_id} / {e}")
            else:
                raise HTTPError(f"Failed to upload reference file by url: {url} / {e}")

    @validate_args(reference_files=Rules.list_not_none_or_none, template_question_id=Rules.uuid_not_none)
    def upload_reference_files_by_url_and_file_paths(self, reference_files: TYPE_OF_REFERENCE_FILES, template_question_id: str) -> None:
        """
        Upload reference files by getting presigned URLs and uploading to them

        Args:
            reference_files: List of reference files with 'path' and 'type' keys
            template_question_id: Template question ID
        """
        if not reference_files:
            return

        try:
            references: List[Dict[Literal["reference_uri", "type", "label_text"], Any]] = []

            # Step 1: Get presigned URLs and upload each file
            for reference_file in reference_files:
                file_path = reference_file["path"]
                file_type = reference_file["type"]

                # Get presigned URL for this file
                response = self.api_client.get(f"template-questions/{template_question_id}/references/presigned-url")
                response.raise_for_status()
                presigned_data = response.json()

                presigned_url = presigned_data["url"]
                reference_uri = presigned_data["uri"]

                # Upload file to presigned URL
                with open(file_path, "rb") as file:
                    upload_response = self.api_client.external_put(presigned_url, data=file)
                    upload_response.raise_for_status()

                # Collect reference data for final submission
                references.append({"reference_uri": reference_uri, "type": file_type, "label_text": None})

                log.debug(f"Successfully uploaded reference file {file_path} for question {template_question_id}")

            # Step 2: Submit all references
            response = self.api_client.put(f"template-questions/{template_question_id}/references", data={"references": references})
            response.raise_for_status()

            log.info(f"Successfully uploaded {len(references)} reference files for question {template_question_id}")

        except Exception as e:
            raise HTTPError(f"Failed to upload reference files for question id: {template_question_id} / {e}")



from typing import Optional

from podonos.core.api import APIClient
from podonos.core.base import log
from podonos.core.config import EvalConfig, EvalConfigDefault
from podonos.core.evaluator import Evaluator

from podonos.common.constant import PODONOS_CONTACT_EMAIL
from podonos.common.enum import AIEvalType, EvalType
from podonos.service.ai_evaluation_service import AIEvaluationService
from podonos.common.validator import Rules, validate_args


class AIEvaluation:
    _api_client: APIClient
    _ai_evaluation_service: AIEvaluationService

    def __init__(self, api_client: APIClient):
        self._api_client = api_client
        self._ai_evaluation_service = AIEvaluationService(api_client)

    @validate_args(
        name=Rules.str_not_none_or_none,
        desc=Rules.str_not_none_or_none,
        type=Rules.str_not_none,
        lan=Rules.str_not_none,
        ai_type=Rules.instance_of(AIEvalType),
        max_upload_workers=Rules.int_not_none,
    )
    def create(
        self,
        name: Optional[str] = None,
        desc: Optional[str] = None,
        type: str = EvalType.CUSTOM_SINGLE.value,
        lan: str = EvalConfigDefault.LAN.value,
        ai_type: AIEvalType = AIEvalType.ALL,
        max_upload_workers: int = EvalConfigDefault.MAX_UPLOAD_WORKERS,
    ) -> Evaluator:
        """
        Create an AI evaluator. For getting ASR results, file's script must be provided.

        Args:
            name: This session name. Its length must be > 1. If empty, a random name is used. Optional.
            desc: Description of this session. Optional.
            type: Only ASR is supported. Default: ASR
            lan: Human language for this audio. One of those in Language. Default: en-us
            ai_type: AI type. Default: ASR
            max_upload_workers: The maximum number of upload workers. Must be a positive integer. Default: 20
        """
        if not AIEvalType.is_ai_type(ai_type):
            raise ValueError(f'"type" must be one of {{ALL}}. \n' + f"Do you want other evaluation types? Let us know at {PODONOS_CONTACT_EMAIL}")

        evaluator = Evaluator(
            self._api_client,
            EvalConfig(name=name, desc=desc, type=type, lan=lan, max_upload_workers=max_upload_workers, ai_type=ai_type),
            [EvalType(type)],
        )
        if ai_type == AIEvalType.ALL:
            log.info(f"Creating AI evaluation for {evaluator.get_evaluation_id()} with {AIEvalType.ALL}")
            self._ai_evaluation_service.create(evaluator.get_evaluation_id(), AIEvalType.ALL)
        return evaluator

    @validate_args(name=Rules.str_not_none_or_none, desc=Rules.str_not_none_or_none, lan=Rules.str_not_none, max_upload_workers=Rules.int_not_none)
    def asr(
        self,
        name: Optional[str] = None,
        desc: Optional[str] = None,
        lan: str = EvalConfigDefault.LAN.value,
        max_upload_workers: int = EvalConfigDefault.MAX_UPLOAD_WORKERS,
    ) -> Evaluator:
        """
        Create an ASR evaluator. file's script must be provided.

        Args:
            name: This session name. Its length must be > 1. If empty, a random name is used. Optional.
            desc: Description of this session. Optional.
            lan: Human language for this audio. One of those in Language. Default: en-us
            max_upload_workers: The maximum number of upload workers. Must be a positive integer. Default: 20
        """
        return self.create(name, desc, EvalType.CUSTOM_SINGLE.value, lan, AIEvalType.ASR, max_upload_workers)



from __future__ import annotations
import inspect
from functools import wraps
import os
from typing import Callable, Any, Tuple, Type, Union, TypeVar
import uuid
from datetime import date, datetime

try:
    from typing import ParamSpec  # Python 3.10+
except Exception:  # pragma: no cover
    from typing_extensions import ParamSpec  # type: ignore

from podonos.core.base import log

P = ParamSpec("P")
R = TypeVar("R")
ValidateFunc = Callable[[Any, str], None]


def validate_args(**validators: Union[Type[Any], Tuple[Type[Any], ...], Callable[[Any, str], None]]) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Decorator to define validation rules for function arguments.

    Args:
        **validators: Maps argument names to validation rules.
                      Rules can be types (str, int, (float, int)) or
                      functions that take (value, name) and perform validation.

    Returns:
        A decorator that preserves the original function's type signature for IDEs.
    """

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:  # type: ignore[misc]
            # 1. Bind arguments: connect passed arguments (args, kwargs) to function signature
            sig = inspect.signature(func)  # type: ignore[arg-type]
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            # 2. Run validation
            for name, rule in validators.items():
                value = bound_args.arguments.get(name)

                # Check if the argument to be tested was actually passed
                if name not in bound_args.arguments:
                    continue  # pass if using default values (can also test default values if needed)

                # A. Type or Not None check (most cases)
                if isinstance(rule, (Type, tuple)):
                    # log.check_notnone raises an exception internally, so we perform None check here instead
                    log.check_notnone(value, f"Argument '{name}' must not be None")  # type: ignore

                    # Type check
                    log.check(isinstance(value, rule), f"Argument '{name}' must be {rule}, got {type(value)}")  # type: ignore

                # B. Custom validation function (complex conditions like length, positive)
                elif callable(rule):
                    # Call rule(value, name) to perform complex validation
                    rule(value, name)

            # 3. Call original function
            return func(*args, **kwargs)

        return wrapper

    return decorator


# -----------------------------
# Centralized Validation Utility
# -----------------------------
class Validator:
    """Centralized reusable validation helpers."""

    # --- Base checks ---
    @staticmethod
    def check_not_none(value: Any, name: str):
        log.check_notnone(value, f"Argument '{name}' must not be None")  # type: ignore

    @staticmethod
    def check_type(value: Any, name: str, expected_type: Union[Type[Any], Tuple[Type[Any], ...]]):
        log.check(isinstance(value, expected_type), f"Argument '{name}' must be {expected_type}, got {type(value)}")  # type: ignore

    @staticmethod
    def check_non_empty_str(value: Any, name: str):
        log.check(isinstance(value, str), f"Argument '{name}' must be a string")  # type: ignore
        log.check(len(value.strip()) > 0, f"Argument '{name}' must not be empty")  # type: ignore

    @staticmethod
    def check_positive(value: Any, name: str):
        log.check(value > 0, f"Argument '{name}' must be positive, got {value}")  # type: ignore

    @staticmethod
    def check_non_empty_list(value: Any, name: str):
        log.check(isinstance(value, list), f"Argument '{name}' must be a list")  # type: ignore
        log.check(len(value) > 0, f"Argument '{name}' must be non-empty list")  # type: ignore

    @staticmethod
    def check_uuid(value: Any, name: str):
        try:
            uuid.UUID(str(value))
        except ValueError:
            raise ValueError(f"Argument '{name}' must be a valid UUID")

    @staticmethod
    def check_instance_of(value: Any, name: str, expected_type: Type[Any]) -> None:
        if not isinstance(value, expected_type):
            raise TypeError(f"{name} must be instance of {expected_type.__name__}, got {type(value).__name__}")

    @staticmethod
    def check_date(value: Any, name: str):
        log.check(isinstance(value, date), f"Argument '{name}' must be a date")  # type: ignore

    @staticmethod
    def check_file(value: Any, name: str):
        log.check(isinstance(value, str), f"Argument '{name}' must be a string")  # type: ignore
        log.check(os.path.isfile(value), f"Argument '{name}' doesn't exist")  # type: ignore
        log.check(os.access(value, os.R_OK), f"Argument '{name}' isn't readable")  # type: ignore


class Rules:
    """Reusable high-level validation rules for @validate_args."""

    # -----------------------------
    # Factory for type-based rules
    # -----------------------------
    @staticmethod
    def make_type_rule(expected_type: Union[Type[Any], Tuple[Type[Any], ...]]) -> ValidateFunc:
        """Return a validator that ensures non-None and type match."""

        def _check(value: Any, name: str) -> None:
            Validator.check_not_none(value, name)
            Validator.check_type(value, name, expected_type)

        return _check

    @staticmethod
    def make_non_empty_str() -> ValidateFunc:
        """Return a validator for non-empty string."""

        def _check(value: Any, name: str) -> None:
            Validator.check_not_none(value, name)
            Validator.check_non_empty_str(value, name)

        return _check

    @staticmethod
    def make_non_empty_list() -> ValidateFunc:
        """Return a validator for non-empty list."""

        def _check(value: Any, name: str) -> None:
            Validator.check_not_none(value, name)
            Validator.check_non_empty_list(value, name)

        return _check

    @staticmethod
    def make_uuid_rule() -> ValidateFunc:
        """Return a validator for valid UUID string."""

        def _check(value: Any, name: str) -> None:
            Validator.check_not_none(value, name)
            Validator.check_uuid(value, name)

        return _check

    @staticmethod
    def make_instance_of_rule(expected_type: Type[Any]) -> ValidateFunc:
        """Return a validator for instance of type."""

        def _check(value: Any, name: str) -> None:
            Validator.check_not_none(value, name)
            Validator.check_instance_of(value, name, expected_type)

        return _check

    @staticmethod
    def make_positive_rule() -> ValidateFunc:
        """Return a validator for positive number."""

        def _check(value: Any, name: str) -> None:
            Validator.check_not_none(value, name)
            Validator.check_type(value, name, (int, float))
            Validator.check_positive(value, name)

        return _check

    @staticmethod
    def make_file_rule() -> ValidateFunc:
        """Return a validator for file."""

        def _check(value: Any, name: str) -> None:
            Validator.check_not_none(value, name)
            Validator.check_file(value, name)

        return _check

    # Declarations for linter/type-checkers (assigned after class definition)
    # Predefined rules (common use)
    str_non_empty: ValidateFunc
    str_not_none: ValidateFunc
    int_not_none: ValidateFunc
    float_not_none: ValidateFunc
    number_not_none: ValidateFunc
    bool_not_none: ValidateFunc
    list_not_none: ValidateFunc
    list_not_empty: ValidateFunc
    uuid_not_none: ValidateFunc
    dict_not_none: ValidateFunc
    tuple_not_none: ValidateFunc
    set_not_none: ValidateFunc
    positive_not_none: ValidateFunc
    date_not_none: ValidateFunc
    datetime_not_none: ValidateFunc
    file_path_not_none: ValidateFunc

    # -----------------------------
    # Instance of (type) variants
    # -----------------------------
    instance_of: Callable[[Type[Any]], ValidateFunc] = make_instance_of_rule

    # -----------------------------
    # Optional (nullable) variants
    # -----------------------------
    @staticmethod
    def optional(expected_type: Union[Type[Any], Tuple[Type[Any], ...]]) -> ValidateFunc:
        """Allow None, otherwise enforce type."""

        def _check(value: Any, name: str) -> None:
            if value is not None:
                Validator.check_type(value, name, expected_type)

        return _check

    @staticmethod
    def optional_uuid(value: Any, name: str) -> None:
        if value is not None:
            Validator.check_uuid(value, name)

    @staticmethod
    def optional_instance_of(expected_type: Type[Any]) -> ValidateFunc:
        """Allow None, otherwise enforce instance of type."""

        def _check(value: Any, name: str) -> None:
            if value is not None:
                Validator.check_instance_of(value, name, expected_type)

        return _check

    # Optional rule declarations (assigned after class definition)
    str_non_empty_or_none: ValidateFunc
    str_not_none_or_none: ValidateFunc
    number_not_none_or_none: ValidateFunc
    list_not_none_or_none: ValidateFunc
    uuid_not_none_or_none: ValidateFunc
    is_optional_instance_of: Callable[[Type[Any]], ValidateFunc]
    datetime_not_none_or_none: ValidateFunc
    dict_not_none_or_none: ValidateFunc


# -----------------------------
# Assign rule callables (3.8/3.9 compatible)
# -----------------------------
# Predefined rules (common use)
Rules.str_non_empty = Rules.make_non_empty_str()
Rules.str_not_none = Rules.make_type_rule(str)
Rules.int_not_none = Rules.make_type_rule(int)
Rules.float_not_none = Rules.make_type_rule(float)
Rules.number_not_none = Rules.make_type_rule((int, float))
Rules.bool_not_none = Rules.make_type_rule(bool)
Rules.list_not_none = Rules.make_type_rule(list)
Rules.list_not_empty = Rules.make_non_empty_list()
Rules.uuid_not_none = Rules.make_uuid_rule()
Rules.dict_not_none = Rules.make_type_rule(dict)
Rules.tuple_not_none = Rules.make_type_rule(tuple)
Rules.set_not_none = Rules.make_type_rule(set)
Rules.positive_not_none = Rules.make_positive_rule()
Rules.date_not_none = Rules.make_type_rule(date)
Rules.datetime_not_none = Rules.make_type_rule(datetime)
Rules.file_path_not_none = Rules.make_file_rule()

# Optional (nullable) variants
Rules.str_non_empty_or_none = Rules.optional(str)
Rules.str_not_none_or_none = Rules.optional(str)
Rules.number_not_none_or_none = Rules.optional((int, float))
Rules.list_not_none_or_none = Rules.optional(list)
Rules.uuid_not_none_or_none = Rules.optional_uuid
Rules.is_optional_instance_of = Rules.optional_instance_of
Rules.datetime_not_none_or_none = Rules.optional(datetime)
Rules.dict_not_none_or_none = Rules.optional(dict)



import hashlib
import json
import os
from typing import Any, Dict, List, Literal, Optional

from requests import Response
from tqdm import tqdm

from podonos.common.constant import CONTENT_TYPE_TO_EXTENSION
from podonos.common.exception import HTTPError
from podonos.common.util import get_content_type_by_filename
from podonos.common.validator import Rules, validate_args
from podonos.core.api import APIClient
from podonos.core.base import log
from podonos.core.config import EvalConfig
from podonos.core.file import Audio, AudioGroup
from podonos.entity.evaluation import EvaluationEntity
from podonos.entity.verification import ProcessFilesResponse, VerifyFilesResponse


class EvaluationService:
    """Service class for handling evaluation-related API communications"""

    def __init__(self, api_client: APIClient):
        self.api_client = api_client

    @validate_args(config=Rules.instance_of(EvalConfig))
    def create(self, config: EvalConfig) -> EvaluationEntity:
        """
        Create a new evaluation based on the evaluation configuration

        Raises:
            HTTPError: If the value is invalid

        Returns:
            Evaluation: Get new evaluation information
        """
        log.debug("Create evaluation")
        try:
            response = self.api_client.post("evaluations", data=config.to_create_request_dto())
            response.raise_for_status()
            evaluation = EvaluationEntity.from_dict(response.json())
            log.info(f"Evaluation is generated: {evaluation.id}")
            return evaluation
        except Exception as e:
            raise HTTPError(f"Failed to create the evaluation: {e}")

    @validate_args(config=Rules.instance_of(EvalConfig))
    def create_from_template(self, config: EvalConfig) -> EvaluationEntity:
        """
        Create a new evaluation based on built-in template

        Raises:
            HTTPError: If the template id is invalid

        Returns:
            Evaluation: Get new evaluation information
        """
        log.debug("Create Evaluation from Template")
        try:
            response = self.api_client.post(
                "evaluations/templates",
                data=config.to_create_from_template_request_dto(),
            )
            response.raise_for_status()
            evaluation = EvaluationEntity.from_dict(response.json())
            log.info(f"Evaluation is generated: {evaluation.id}")
            return evaluation
        except Exception as e:
            raise HTTPError(f"Failed to create the evaluation: {e}")

    @validate_args(evaluation_id=Rules.uuid_not_none)
    def get_evaluation(self, evaluation_id: str) -> EvaluationEntity:
        """Get evaluation by ID"""
        try:
            response = self.api_client.get(f"evaluations/{evaluation_id}")
            response.raise_for_status()
            return EvaluationEntity.from_dict(response.json())
        except Exception as e:
            raise HTTPError(f"Failed to get evaluation: {e}")

    def get_evaluation_list(self) -> List[Dict[str, Any]]:
        """Gets a list of evaluations.

        Args: None

        Returns:
            Evaluation containing all the evaluation info
        """
        try:
            response = self.api_client.get("evaluations")
            response.raise_for_status()
            evaluations = [EvaluationEntity.from_dict(evaluation) for evaluation in response.json()]
            return [evaluation.to_dict() for evaluation in evaluations]
        except Exception as e:
            raise HTTPError(f"Failed to get evaluation list: {e}")

    @validate_args(evaluation_id=Rules.uuid_not_none, group_by=Rules.str_not_none)
    def get_stats_json_by_id(
        self,
        evaluation_id: str,
        group_by: Literal["question", "script", "model"] = "question",
    ) -> List[Dict[str, Any]]:
        """Gets a list of evaluation statistics referenced by id.

        Args:
            evaluation_id: Evaluation id. See get_evaluation_list() above.
            group_by: Group by question or script. Default: "question". "script" is only available for single-question evaluation.

        Returns:
            List of statistics for the evaluation.
        """
        try:
            response = self.api_client.get(f"evaluations/{evaluation_id}/stats?group-by={group_by}")
            if response.status_code == 400:
                log.info(f"Bad Request: The {evaluation_id} is an invalid evaluation id")
                return []

            response.raise_for_status()
            return response.json()
        except Exception as e:
            raise HTTPError(f"Failed to get evaluation stats: {e}")

    @validate_args(evaluation_id=Rules.uuid_not_none, audios=Rules.list_not_none)
    def create_evaluation_files(self, evaluation_id: str, audios: List[Audio]):
        try:
            response = self.api_client.put(
                f"evaluations/{evaluation_id}/files",
                {"files": [audio.to_create_file_dict() for audio in audios]},
            )
            response.raise_for_status()
        except Exception as e:
            log.error(f"HTTP error in adding file meta: {e}")
            raise HTTPError(
                f"Failed to create evaluation files: {e}",
                status_code=getattr(getattr(e, "response", None), "status_code", None),
            )

    @validate_args(evaluation_id=Rules.uuid_not_none, remote_object_name=Rules.str_not_none)
    def get_presigned_url(self, evaluation_id: str, remote_object_name: str) -> str:
        """Get presigned URL for file upload"""
        try:
            response = self.api_client.put(
                f"evaluations/{evaluation_id}/uploading-presigned-url",
                data={"uploaded_file_name": remote_object_name},
            )
            response.raise_for_status()
            return response.text.replace('"', "")
        except Exception as e:
            log.error(f"HTTP error in getting a presigned url: {e}")
            raise HTTPError(f"Failed to get presigned URL: {e}")

    @validate_args(url=Rules.str_not_none, path=Rules.file_path_not_none)
    def upload_evaluation_file(self, url: str, path: str) -> Response:
        try:
            with open(path, "rb") as file:
                response = self.api_client.external_put(
                    url,
                    data=file,
                    headers={"Content-Type": get_content_type_by_filename(path)},
                )
            return response
        except Exception as e:
            log.error(f"HTTP error in uploading a file to presigned URL: {e}")
            raise HTTPError(
                f"Failed to Upload File {path}: {e}",
                status_code=getattr(getattr(e, "response", None), "status_code", None),
            )

    @validate_args(
        evaluation_id=Rules.uuid_not_none,
        config=Rules.instance_of(EvalConfig),
        audio_groups=Rules.list_not_none,
    )
    def upload_session_json(self, evaluation_id: str, config: EvalConfig, audio_groups: List[AudioGroup]) -> None:
        """Upload session JSON data"""
        try:
            session_json = config.to_dict()
            session_json["files"] = [group.to_dict() for group in audio_groups]
            presigned_url = self.get_presigned_url(evaluation_id, "session.json")
            self.put_session_json(
                presigned_url,
                session_json,
                headers={"Content-type": "application/json"},
            )
        except Exception as e:
            raise HTTPError(f"Failed to upload session JSON: {e}")

    @validate_args(
        url=Rules.str_not_none,
        data=Rules.dict_not_none,
        headers=Rules.dict_not_none_or_none,
    )
    def put_session_json(self, url: str, data: Dict[str, Any], headers: Optional[Dict[str, str]] = None) -> Response:
        log.debug("JSON data")
        for key, value in data.items():
            log.debug(f"{key}: {value}")
        if headers:
            log.debug("Headers")
            for key, value in headers.items():
                log.debug(f"{key}: {value}")

        try:
            response = self.api_client.external_put(url, json_data=data, headers=headers)
            return response
        except Exception as e:
            log.error(f"HTTP error in uploading a json to presigned url: {e}")
            raise HTTPError(
                f"Failed to Upload JSON {data}: {e}",
                status_code=getattr(getattr(e, "response", None), "status_code", None),
            )

    @validate_args(evaluation_id=Rules.uuid_not_none, output_dir=Rules.str_not_none)
    def download_evaluation_files_by_evaluation_id(self, evaluation_id: str, output_dir: str) -> str:
        """Download evaluation files using CloudFront cookies."""
        try:
            # Get the response from the API
            log.debug(f"Download evaluation files for evaluation {evaluation_id}")
            file_mata_json: Dict[str, List[Dict[str, Any]]] = {"files": []}
            response = self.api_client.get(f"evaluation-files/download?evaluation-id={evaluation_id}")
            response.raise_for_status()

            # Parse the response using EvaluationFileDownloadResponseDto
            download_response = response.json()

            # Ensure the output directory exists
            os.makedirs(output_dir, exist_ok=True)

            # Download each file using the original URL and cookies
            for file in tqdm(download_response["files"], desc="Downloading files", unit="file"):
                # Download the file using the original URL and cookies
                file_response = self.api_client.external_get(file["original_url"], cookies=download_response["cookie"])
                file_response.raise_for_status()

                content_type = file_response.headers.get("Content-Type")
                file_extension = CONTENT_TYPE_TO_EXTENSION[content_type] if content_type else ".flac"

                # Generate a hash for the original file name
                file_original_name = file["original_name"]
                hash_object = hashlib.md5(file_original_name.encode(), usedforsecurity=False)
                hashed_file_name = hash_object.hexdigest()

                # Construct the file path using the model tag and hashed file name
                file_name = f"{file['model_tag']}/{hashed_file_name}{file_extension}"
                file_path = os.path.join(output_dir, file_name)

                # Ensure the directory for the model tag exists
                os.makedirs(os.path.dirname(file_path), exist_ok=True)

                # Save the file locally
                with open(file_path, "wb") as f:
                    f.write(file_response.content)

                file_mata_json["files"].append(
                    {
                        "file_path": file_path,
                        "original_name": file_original_name,
                        "model_tag": file["model_tag"],
                        "tags": file["tags"],
                    }
                )

            log.info(f"Downloaded {len(file_mata_json['files'])} files")

            # Save the file metadata to a JSON file
            metadata_file_path = os.path.join(output_dir, "metadata.json")
            with open(metadata_file_path, "w") as f:
                json.dump(file_mata_json, f)
            log.info(f"File metadata saved to {metadata_file_path}")

            return "Files downloaded successfully."
        except Exception as e:
            raise HTTPError(f"Failed to download evaluation files: {e}")

    @validate_args(evaluation_id=Rules.uuid_not_none, audios=Rules.list_not_none)
    def verify_files(self, evaluation_id: str, audios: List[Audio]) -> VerifyFilesResponse:
        try:
            payload = {
                "files": [
                    {
                        "uploaded_file_name": audio.remote_object_name,
                        "content_md5": audio.content_md5,
                        "file_size": audio.file_size,
                    }
                    for audio in audios
                ]
            }
            response = self.api_client.post(f"evaluations/{evaluation_id}/files/verify", data=payload)
            response.raise_for_status()
            return VerifyFilesResponse.from_dict(response.json())
        except Exception as e:
            raise HTTPError(
                f"Failed to verify evaluation files: {e}",
                status_code=getattr(getattr(e, "response", None), "status_code", None),
            )

    @validate_args(evaluation_id=Rules.uuid_not_none, file_meta_ids=Rules.list_not_none_or_none)
    def process_files(self, evaluation_id: str, file_meta_ids: Optional[List[str]] = None) -> ProcessFilesResponse:
        try:
            payload: Dict[str, Any] = {"file_meta_ids": file_meta_ids} if file_meta_ids else {}
            response = self.api_client.post(f"evaluations/{evaluation_id}/files/process", data=payload)
            response.raise_for_status()
            return ProcessFilesResponse.from_dict(response.json())
        except Exception as e:
            raise HTTPError(
                f"Failed to trigger file processing: {e}",
                status_code=getattr(getattr(e, "response", None), "status_code", None),
            )



from podonos.errors.error import (
    FileVerificationFailure,
    InvalidFileError,
    NotSupportedError,
    UploadRetryExhaustedError,
    UploadVerificationError,
)

__all__ = [
    "FileVerificationFailure",
    "InvalidFileError",
    "NotSupportedError",
    "UploadRetryExhaustedError",
    "UploadVerificationError",
]
